"""Constants used by tests"""

TEST_V4_PATH = "/v4/timelines"

REALTIME_FIELDS_GREATER_THAN_MAX = [
    "temperature",
    "pressureSeaLevel",
    "humidity",
    "windSpeed",
    "windDirection",
    "pollutantO3",
    "weatherCode",
    "visibility",
    "windGust",
    "cloudCover",
    "precipitationType",
    "particulateMatter25",
    "particulateMatter10",
    "pollutantNO2",
    "pollutantCO",
    "pollutantSO2",
    "uvIndex",
    "uvHealthConcern",
    "epaIndex",
    "epaPrimaryPollutant",
    "epaHealthConcern",
    "mepIndex",
    "mepPrimaryPollutant",
    "mepHealthConcern",
    "treeIndex",
    "treeAcaciaIndex",
    "treeAshIndex",
    "treeBeechIndex",
    "treeBirchIndex",
    "treeCedarIndex",
    "treeCypressIndex",
    "treeElderIndex",
    "treeElmIndex",
    "treeHemlockIndex",
    "treeHickoryIndex",
    "treeJuniperIndex",
    "treeMahoganyIndex",
    "treeMapleIndex",
    "treeMulberryIndex",
    "treeOakIndex",
    "treePineIndex",
    "treeCottonwoodIndex",
    "treeSpruceIndex",
    "treeSycamoreIndex",
    "treeWalnutIndex",
    "treeWillowIndex",
    "weedIndex",
    "weedRagweedIndex",
    "grassIndex",
    "grassGrassIndex",
    "fireIndex",
    "temperatureApparent",
    "dewPoint",
    "pressureSurfaceLevel",
    "solarGHI",
    "cloudBase",
    "cloudCeiling",
]

CORE_FIELDS = [
    "temperature",
    "temperatureApparent",
    "dewPoint",
    "humidity",
    "windSpeed",
    "windGust",
    "pressureSurfaceLevel",
    "pressureSeaLevel",
    "precipitationIntensity",
    "rainIntensity",
    "freezingRainIntensity",
    "snowIntensity",
    "sleetIntensity",
    "precipitationProbability",
    "rainAccumulation",
    "snowAccumulation",
    "sleetAccumulation",
    "iceAccumulation",
    "visibility",
    "cloudCover",
    "cloudBase",
    "cloudCeiling",
]
